(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2009-15 Univ. Paris-Sud, France
 *               2009-15 Achim D. Brucker, Germany
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)

theory  MyKeOS_test_conc_opt
imports 
   (* Testing *) 
   (*"../../../src/codegen_fsharp/Code_Integer_Fsharp"*) 
   MyKeOS
   POR_Testing
begin


section{* Partial Order Reduction Approach -- Old Testament by bu*}


text{* The continuation monad $M$ is conceptually superfluous; it is used 
       to carry the result type-parameter @{typ "'o'"} to maintain full
       generality. This can not be captured by a universal quantification \emph{inside}
       the definition. *}
definition po_is_exec_consistent :: 
           "('\<iota> \<Rightarrow> ('o,'\<sigma>)MON_SE) \<Rightarrow> ('\<iota> \<Rightarrow> '\<iota> \<Rightarrow>  bool) \<Rightarrow>  ('o \<Rightarrow> 'o \<Rightarrow> ('o','\<sigma>)MON_SE) \<Rightarrow> bool"
where      "po_is_exec_consistent iostep ord M =            
                (\<forall> \<iota> \<iota>'. 
                    (ord \<iota>' \<iota>) \<longrightarrow>                
                         ((o1 \<leftarrow> iostep \<iota>;  (o2 \<leftarrow> iostep \<iota>'; M o1 o2))) = 
                         ((o2 \<leftarrow> iostep \<iota>'; (o1 \<leftarrow> iostep \<iota>;  M o1 o2))))"



section{* Commutativities for Re-Orderding Theorems and Partial Order Reduction*}

text{* This ordering has the intuition: "syntactic smaller"; it is later on shown
       that two inputs are only put into this ordering if the operations commute.
       We will define interference in terms of this partial order.
     *}


instantiation in_c::ord
begin

fun  less_in_c where     
    "(alloc c no m)   < (alloc c' no' m')   = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(alloc c no m)   < (release c' no' m') = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(alloc c no m)   < (status c' no')     = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(release c no m) < (release c' no' m') = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(release c no m) < (status c' no')     = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(status c no )   < (status c' no' )    = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(X::in_c)        < (Y::in_c)           = False"


definition less_eq_in_c_def : "(s::in_c) \<le> t \<longleftrightarrow> s < t \<or> s = t" 

instance ..

end



lemma less_in_c_nonrefl : "(X::in_c) < (Y::in_c) \<Longrightarrow> X \<noteq> Y"
by(cases "X")auto

lemma less_in_c_antisym : "(X::in_c) < (Y::in_c) \<Longrightarrow> \<not>(Y < X)"
apply(cases "X")
apply(cases "Y",auto)+
done


instantiation in_c::order
begin
instance
proof
  fix x::in_c show "x \<le> x"  by(simp add: less_eq_in_c_def) 
next
  fix x::in_c fix y assume "x \<le> y" and "y \<le> x" thus  " x = y"  
      apply(auto simp: less_eq_in_c_def, cases "x")
      apply(cases "y",auto)+
      done
next
  fix x::in_c fix y  fix z show "x \<le> y \<Longrightarrow> y \<le> z \<Longrightarrow> x \<le> z" 
      apply(auto simp: less_eq_in_c_def)
      apply(cases "y", simp_all)
      apply(cases "x",cases "z",auto)
      apply(cases "x",cases "z",auto)
      apply(cases "z", auto) apply(cases "z", auto)
      apply(cases "x",cases "z",auto)
      apply(cases "z", auto) +
      done
next
  fix x::in_c fix y show  "(x < y) = (x \<le> y \<and> \<not> y \<le> x)"  
      apply(auto simp: less_eq_in_c_def dest:less_in_c_nonrefl)
      apply(drule less_in_c_antisym, simp)
      done
qed

end


lemma "(x::in_c) \<le> y \<or> y \<le> x"
(* But: order not antisymmetric !!! This here is simply not True ! *)
oops
(* Therefore: This is impossible ...
instantiation in_c::linorder
begin
instance
proof
   fix x::in_c and y::in_c show "x \<le> y \<or> y \<le> x"
      apply(auto simp: less_eq_in_c_def)
thm less_in_c_antisym
      sorry 
qed
*)



definition interfer :: "in_c \<Rightarrow> in_c \<Rightarrow> bool" (infixl "\<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s" 55)
where "x \<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s y = (\<not> x \<le> y \<and> \<not> y \<le> x )"


lemma interfer_irrefl : "\<not>(x \<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s x)"
unfolding interfer_def
by auto


lemma interfer_sym: "x \<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s y \<Longrightarrow> y \<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s x"
unfolding interfer_def
by auto

lemma interfer_sym: "x \<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s y \<Longrightarrow> y \<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s z \<Longrightarrow> x \<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s z"
txt{* Interference is not transitive ... not a real surprise ... *}
oops 

lemma thibo1 : "(alloc c no m)  \<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s (status c no)"
unfolding interfer_def less_eq_in_c_def
by auto

lemma thibo2 : "(status c no)  \<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s (alloc c' no' m) = (c = c' \<and> no = no')"
unfolding interfer_def less_eq_in_c_def
oops

subsection{* Permutations and a specification of consistent reorderings *}
 



text{* Now comes a specification of a correct re-orderering function *}
definition reorder_correct 
where     "reorder_correct reorder = 
                (\<forall>S. \<exists>index_perm. is_perm [0 ..< length S] index_perm \<and>
                                  (reorder S) = map (op ! S) index_perm \<and> 
                                  (\<forall>k < length S.  \<forall>i < k.  (S!i) \<approx>\<^sub>i\<^sub>n\<^sub>t\<^sub>e\<^sub>r\<^sub>f\<^sub>e\<^sub>r\<^sub>s (S!k) \<longrightarrow> 
                                                            index_perm ! i < index_perm ! k))"

(* yes but ... even the identity is a correct reorder*)
lemma "reorder_correct (\<lambda>x. x)"
unfolding reorder_correct_def
by(auto,rule_tac x="[0..<length S]" in exI,simp add: List.map_nth)



value "(alloc 1 1 10)   < (alloc 1 2 5)"
value "(alloc 1 2 5)    < (release 1 1 7)"
value "(alloc 1 1 5)    < (release 1 1 7)"
value "(release 1 1 5)  < (alloc 1 1 7)"
value "(release 1 2 5)  < (release 1 1 7)"
value "(release 1 1 7)  < (alloc 1 2 5)"

text{* Observation : we are actually \emph{not} interested in topsorting; rather, we are
       interested in re-ordering as efficiently as possible the various interleavings.
       Optimally, we would target at a \emph{canonizing} re-order, but we could
       do without.
        *}

fun    select_smallest\<^sub>s\<^sub>t\<^sub>d :: "'a\<Colon>order \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a \<times> 'a list \<times> 'a list"
where "select_smallest\<^sub>s\<^sub>t\<^sub>d a [] T U = (a,rev T,rev U)"
     |"select_smallest\<^sub>s\<^sub>t\<^sub>d a (b#S) T U = (if a \<le> b then  select_smallest\<^sub>s\<^sub>t\<^sub>d a S (b#T) U
                                        else if b \<le> a then select_smallest\<^sub>s\<^sub>t\<^sub>d b S (a#T) U  (* really ??? *)
                                             else select_smallest\<^sub>s\<^sub>t\<^sub>d a S T (b#U))"

value "select_smallest\<^sub>s\<^sub>t\<^sub>d (alloc 1 1 10)[release 1 1 7, alloc 1 2 5, release 1 2 3,  status 1 1] [] [] "

fun    select_smallest :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a \<times> 'a list \<times> 'a list"
where "select_smallest ord a [] T U = (a,rev T,rev U)"
     |"select_smallest ord a (b#S) T U = (if ord a b then select_smallest ord a S (b#T) U
                                     else if ord b a then select_smallest ord b S (a#T) U  (* really ??? *)
                                                     else select_smallest ord a S T (b#U))"


lemma select_smallest\<^sub>s\<^sub>t\<^sub>d_inv [simp, rule_format] : 
"(x, T', U') = select_smallest\<^sub>s\<^sub>t\<^sub>d a S T U \<longrightarrow> length T' + length U' = length S + length T + length U"
proof(induct S arbitrary: a T U)
    case Nil        show ?case by simp
next
    case (Cons a S) from Cons.hyps show ?case by auto
qed

lemma select_smallest_inv [simp, rule_format] : 
"(x, T', U') = select_smallest f a S T U \<longrightarrow> length T' + length U' = length S + length T + length U"
proof(induct S arbitrary: a T U)
    case Nil        show ?case by simp
next
    case (Cons a S) from Cons.hyps show ?case by auto
qed

lemma [simp, rule_format] : "(x, T, U) = select_smallest\<^sub>s\<^sub>t\<^sub>d a S [] [] \<longrightarrow> length T \<le> (length S)"
by(auto dest: select_smallest\<^sub>s\<^sub>t\<^sub>d_inv)

lemma [simp, rule_format] : "(x, T, U) = select_smallest f a S [] [] \<longrightarrow> length T \<le> (length S)"
by(auto dest: select_smallest_inv)

fun    reorder\<^sub>s\<^sub>t\<^sub>d :: "'a::order list \<Rightarrow> 'a list"
where "reorder\<^sub>s\<^sub>t\<^sub>d [] = []"
     |"reorder\<^sub>s\<^sub>t\<^sub>d (a#S) = (case select_smallest\<^sub>s\<^sub>t\<^sub>d a S [] [] of
                                (a', I, Dep) \<Rightarrow>  (a' # Dep) @ (reorder\<^sub>s\<^sub>t\<^sub>d I) )"


fun    reorder :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a list"
where "reorder ord [] = []"
     |"reorder ord (a#S) = (case select_smallest ord a S [] [] of
                                (a', I, Dep) \<Rightarrow>  (a' # Dep) @ (reorder ord I) )"




lemma "reorder_correct reorder\<^sub>s\<^sub>t\<^sub>d"
sorry
 

value "reorder\<^sub>s\<^sub>t\<^sub>d [alloc 1 1 10, alloc 1 2 5, release 1 1 7, release 1 2 3,  status 1 1]"
value "card  (interleave (syscall 2 0 1 2) (syscall 2 1 3 4))"
value "card(reorder\<^sub>s\<^sub>t\<^sub>d ` (interleave (syscall 2 0 1 2) (syscall 2 1 3 4)))"

(* Watch out : duplicates have to be erased explicitely from the set representation ... 
   A better solution is this construction on lists ... *)
value "(reorder\<^sub>s\<^sub>t\<^sub>d ` (interleave (syscall 2 0 1 2) (syscall 2 1 3 4)))"
value "remdups(map reorder\<^sub>s\<^sub>t\<^sub>d (interleave\<^sub>l\<^sub>i\<^sub>s\<^sub>t (syscall 2 0 1 2) (syscall 2 1 3 4) ))"

lemma 11:  "((o1 \<leftarrow> SYS (alloc c no m);    (o2 \<leftarrow> SYS (alloc c' no' m');M o1 o2))) = 
            ((o2 \<leftarrow> SYS (alloc c' no' m'); (o1 \<leftarrow> SYS (alloc c no m);   M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def alloc.exec_succes alloc.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_ac)
done

lemma 12:  "(c, no) \<noteq> (c', no') \<Longrightarrow>
            ((o1 \<leftarrow> SYS (alloc c no m);      (o2 \<leftarrow> SYS (release c' no' m'); M o1 o2))) = 
            ((o2 \<leftarrow> SYS (release c' no' m'); (o1 \<leftarrow> SYS (alloc c no m);      M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def alloc.exec_succes alloc.exec_failure release.exec_succes 
                release.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_ac)
done

lemma 22: "((o1 \<leftarrow> SYS (release c no m);    (o2 \<leftarrow> SYS (release c' no' m'); M o1 o2))) = 
           ((o2 \<leftarrow> SYS (release c' no' m'); (o1 \<leftarrow> SYS (release c no m);    M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def release.exec_succes release.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_commute)
done


lemma 33: "((o1 \<leftarrow> SYS (status c no);   (o2 \<leftarrow> SYS (status c' no'); M o1 o2))) = 
           ((o2 \<leftarrow> SYS (status c' no'); (o1 \<leftarrow> SYS (status c no);   M o1 o2)))"
by(rule ext,simp add: bind_SE_def status.exec_succes status.exec_failure split:option.split)

lemma 31: "(c, no) \<noteq> (c', no') \<Longrightarrow>
            ((o1 \<leftarrow> SYS (status c no);  (o2 \<leftarrow> SYS (alloc c' no' m'); M o1 o2))) = 
            ((o2 \<leftarrow> SYS (alloc c' no' m'); (o1 \<leftarrow> SYS (status c no);  M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def alloc.exec_succes alloc.exec_failure status.exec_succes 
                status.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_ac)
done

lemma 32:  "(c, no) \<noteq> (c', no') \<Longrightarrow>
            ((o1 \<leftarrow> SYS (status c no);  (o2 \<leftarrow> SYS (release c' no' m'); M o1 o2))) = 
            ((o2 \<leftarrow> SYS (release c' no' m'); (o1 \<leftarrow> SYS (status c no);  M o1 o2)))"
apply(rule ext)
apply(simp add: bind_SE_def release.exec_succes release.exec_failure status.exec_succes 
                status.exec_failure split:option.split)
apply(auto simp: Fun.fun_upd_twist Groups.ab_group_add_class.diff_diff_eq 
                 Groups.ab_semigroup_add_class.add_ac)
done


theorem SYS_is_exec_consistent : "po_is_exec_consistent SYS (op <) M"
apply(auto simp: po_is_exec_consistent_def)
apply(rename_tac "\<iota>'" "\<iota>")

apply(case_tac "\<iota>'", auto)
  apply(case_tac "\<iota>", auto, subst 11, simp, subst 11, simp)
 apply(case_tac "\<iota>", auto, subst 12, simp_all,subst 12, simp_all)
 apply(subst 22, simp_all, subst 22, simp_all)
apply(case_tac "\<iota>", auto)
    apply(subst 31, simp_all,subst 31, simp_all)
   apply(subst 32, simp_all,subst 32, simp_all)
  apply(subst 33, simp_all)+
done


theorem ordered_exec_state_invariant' :
assumes A: "\<forall>M. po_is_exec_consistent iostep (op <) M"
shows      "(\<sigma> \<Turnstile> (_ \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (S) iostep); assert (P))) = 
            (\<sigma> \<Turnstile> (_ \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p (reorder\<^sub>s\<^sub>t\<^sub>d S) iostep); assert (P)))" 
sorry

term "reorder (\<lambda>(x,y). \<lambda>(x',y'). x < x') (zip S [0 ..< length S])"

theorem ordered_exec_state_invariant'' :
assumes A: "\<forall>M. po_is_exec_consistent iostep (op <) M"
shows      "(\<sigma> \<Turnstile> (os \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p S iostep); M S os)) =
            (let S' = reorder (\<lambda>(x,y). \<lambda>(x',y'). x < x') (zip S [0 ..< length S]);
                 input_perm = map fst S';
                 index_perm = map snd S'
             in  \<sigma> \<Turnstile> (os \<leftarrow> (mbind\<^sub>F\<^sub>a\<^sub>i\<^sub>l\<^sub>S\<^sub>t\<^sub>o\<^sub>p input_perm iostep); 
                      M input_perm (map (op ! os) index_perm)))" 
sorry


subsection{*Partial order reduction with interference as a primitive --- New testament*}
(* by bu + thibo *)


fun interfer_in_c :: "in_c \<Rightarrow> in_c \<Rightarrow> bool" (infix "##\<^sub>c" 65)
where 
  "alloc c no m ##\<^sub>c alloc c' no' m'     = (c=c' \<and> no=no')"
| "alloc c no m ##\<^sub>c release c' no' m'   = (c>c' \<or> c=c' \<and> no\<ge>no')"
| "alloc c no m ##\<^sub>c status c' no'       = (c>c' \<or> c=c' \<and> no\<ge>no')"
| "release c no m ##\<^sub>c alloc c' no' m'   = (c<c' \<or> c=c' \<and> no\<le>no')"
| "release c no m ##\<^sub>c release c' no' m' = (c=c' \<and> no=no')"
| "release c no m ##\<^sub>c status c' no'     = (c>c' \<or> c=c' \<and> no\<ge>no')"
| "status c no ##\<^sub>c alloc c' no' m'      = (c<c' \<or> c=c' \<and> no\<le>no')"
| "status c no ##\<^sub>c release c' no' m'    = (c<c' \<or> c=c' \<and> no\<le>no')"
| "status c no ##\<^sub>c status c' no'        = (c=c' \<and> no=no')"

interpretation itrf_bu : interference_relation "op ##\<^sub>c" "SYS" 

proof
  fix x y::in_c assume "x ##\<^sub>c y" thus "y ##\<^sub>c x"
  apply(cases "x")
  apply(cases "y", auto)+
  done
qed

fun itrf_in_c :: "in_c \<Rightarrow> in_c \<Rightarrow> bool" (infix "##\<^sub>t" 65)
where
  "alloc c no m ##\<^sub>t release c' no' m'   = ((c,no) = (c',no'))"
| "release c' no' m' ##\<^sub>t alloc c no m   = ((c,no) = (c',no'))"
| "alloc c no m ##\<^sub>t status c' no'       = ((c,no) = (c',no'))"
| "status c no ##\<^sub>t alloc c' no' m'      = ((c,no) = (c',no'))"
| "release c no m ##\<^sub>t status c' no'     = ((c,no) = (c',no'))"
| "status c no ##\<^sub>t release c' no' m'    = ((c,no) = (c',no'))"
| "_ ##\<^sub>t _                              = False"

interpretation itrf_bis : interference_relation "op ##\<^sub>t" "SYS"

proof
  fix x y::in_c assume "x ##\<^sub>t y" thus "y ##\<^sub>t x"
  apply(cases "x")
  apply(cases "y", auto)+
  done
qed


(* The following relation should be a standardization order for ##\<^sub>t *)
fun  less_in_c_bis :: "in_c \<Rightarrow> in_c \<Rightarrow> bool" (infix "<\<^sub>t" 65)
where     
    "(alloc c no m)   <\<^sub>t (alloc c' no' m')   = ((c < c') \<or> (c = c' \<and> no < no')
                                                \<or> (c = c' \<and> no = no' \<and> m < m'))"
 |  "(alloc c no m)   <\<^sub>t (release c' no' m') = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(alloc c no m)   <\<^sub>t (status c' no')     = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(release c no m) <\<^sub>t (alloc c' no' m')   = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(release c no m) <\<^sub>t (release c' no' m') = ((c < c') \<or> (c = c' \<and> no < no')
                                                \<or> (c = c' \<and> no = no' \<and> m < m'))"
 |  "(release c no m) <\<^sub>t (status c' no')     = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(status c no)    <\<^sub>t (alloc c' no' m')   = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(status c no)    <\<^sub>t (release c' no' m') = ((c < c') \<or> (c = c' \<and> no < no'))"
 |  "(status c no )   <\<^sub>t (status c' no' )    = ((c < c') \<or> (c = c' \<and> no < no'))"


(* Reflexivity is for this particular case *)
lemma interfer_in_c_refl: "(x::in_c) ##\<^sub>c x"
by(cases "x", simp_all)

theorem SYS_itrf_is_exec_consistent : "itrf_bu.interference_is_exec_consistent M"
apply(auto simp: itrf_bu.interference_is_exec_consistent_def)
apply(rename_tac "\<iota>'" "\<iota>")

apply(case_tac "\<iota>'", auto)
  apply(case_tac "\<iota>", auto, subst 11, simp, subst 11, simp)
     apply(subst 12,simp_all)+
  apply(subst 31,simp_all)+
 apply(case_tac "\<iota>", auto, subst 12, simp_all,subst 12, simp_all)
 apply(subst 22, simp_all, subst 22, simp_all)
   apply(subst 32, simp_all,subst 32, simp_all)
 apply(case_tac "\<iota>", auto)
     apply(subst 31,simp_all)+
apply(subst 32, simp_all)+
  apply(subst 33, simp_all)+
done



interpretation reorder_bu: standardization_order "op \<le>" "op <" "op ##\<^sub>c" "SYS"
proof
  fix x y::in_c assume "x##\<^sub>cy" thus "\<not> x < y"
  apply(cases "x")
  apply(cases "y", auto)+
  done
next
  fix x y::in_c assume "\<not>x##\<^sub>cy" thus "x<y \<or> y<x"
  apply(cases "x")
  apply(cases "y", simp_all, auto)+
  done
qed


end